from dataclasses import dataclass, field
from email.message import EmailMessage
from typing import ClassVar, Iterable, Optional
import mimetypes

from aiosmtplib import send  # type: ignore

from ...port.out_.email_sender import EmailSenderProtocol


@dataclass(slots=True, kw_only=True)
class Attachment:
    content_type: str = field(init=False)
    filename: str
    bytes: bytes

    def __post_init__(self):
        content_type, encoding = mimetypes.guess_type(self.filename)
        if content_type is None or encoding is not None:
            content_type = "application/octet-stream"
        self.content_type = content_type


def build_email_message(
    *,
    from_: str,
    to: str | Iterable[str],
    subject: str,
    text_version: str,
    html_version: Optional[str] = None,
    attachments: Optional[Iterable[Attachment]] = None,
):
    msg = EmailMessage()
    # header
    msg.add_header("From", from_)
    msg.add_header("To", to if isinstance(to, str) else ", ".join(to))
    msg.add_header("Subject", subject)
    # payload
    msg.set_content(text_version)
    if html_version:
        msg.add_alternative(html_version, subtype="html")
    if attachments:
        for attachment in attachments:
            main, sub = attachment.content_type.split("/", 1)
            msg.add_attachment(
                attachment.bytes,
                maintype=main,
                subtype=sub,
                filename=attachment.filename,
            )
    return msg


class BaseEmailNotification(EmailSenderProtocol):

    """
    # Example

    ```python
    class NaverEmailNotification(BaseEmailNotification):
        HOST = "smtp.naver.com"
        PORT = 587
        USERNAME = "naver account id"
        PASSWORD = "naver account password"

    msg = build_email_message(
        from_="from@example.com",
        to=["to@example.com"],
        subject="제목",
        text_version="Text Version",
        html_version="<html>HTML Version</html>",
        attachments=[
            Attachment(filename="filename.txt", bytes=b"file.read()"),
        ],
    )

    sender = NaverEmailNotification()
    asyncio.run(sender.send(msg))
    ```
    """

    HOST: ClassVar[str]
    PORT: ClassVar[int]
    USERNAME: ClassVar[str]
    PASSWORD: ClassVar[str]

    async def send(self, msg: EmailMessage):
        await send(
            msg,
            hostname=self.HOST,
            port=self.PORT,
            username=self.USERNAME,
            password=self.PASSWORD,
            start_tls=True,
        )



from typing import Protocol, TypeVar


EventType = TypeVar("EventType", contravariant=True)


class OutboxProtocol(Protocol[EventType]):
    async def put(self, _event: EventType) -> None:
        ...

    async def delete(self, _event: EventType) -> None:
        ...



from dataclasses import dataclass, field
from types import TracebackType
from typing import Callable, Optional

from sqlalchemy.ext.asyncio import AsyncSession  # type: ignore

from ....port.out_.unit_of_work import UnitOfWorkProtocol


@dataclass
class BaseSQLAlchemyUnitOfWork(UnitOfWorkProtocol):

    SESSION_FACTORY: Callable[[], AsyncSession]

    _session: AsyncSession = field(init=False)

    async def __aenter__(self):
        self._session = self.SESSION_FACTORY()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        await self._session.close()  # type: ignore

    async def commit(self):
        await self._session.commit()  # type: ignore

    async def rollback(self):
        await self._session.rollback()  # type: ignore



import asyncio
from typing import (
    Any,
    Optional,
    Awaitable,
    Callable,
    Literal,
    TypeVar,
    overload,
)

from ..messages.base import Message, MessageCatchContext, get_issued_messages

from .handler import Handler

M = TypeVar("M", bound=Message)
R = TypeVar("R")
Handlers = tuple[Handler[M, R], ...]


Hook = Callable[[M, Handler[M, R]], Awaitable[None]]


async def handle(
    message: M,
    handler: Handler[M, Any],
    deps: dict[str, Any],
    pre_hook: Optional[Hook[M, Any]] = None,
    post_hook: Optional[Hook[M, Any]] = None,
):
    if pre_hook:
        await pre_hook(message, handler)
    result = await handler(message, **deps)
    if post_hook:
        await post_hook(message, handler)
    return result


async def handle_parallel(
    message: M,
    handlers: tuple[Handler[M, Any], ...],
    deps: dict[str, Any],
    pre_hook: Optional[Hook[M, Any]] = None,
    post_hook: Optional[Hook[M, Any]] = None,
):
    coros = (
        handle(message, handler, deps, pre_hook, post_hook) for handler in handlers
    )
    return await asyncio.gather(*coros, return_exceptions=True)


class MessageBus:
    def __init__(
        self,
        *,
        deps: dict[str, Any],
        pre_hook: Optional[Hook[Message, Any]] = None,
        post_hook: Optional[Hook[Message, Any]] = None,
    ) -> None:
        self._deps: dict[str, Any] = deps
        self._handler_map: dict[
            type[Message], Handler[Message, Any] | Handlers[Message, Any]
        ] = {}
        self._pre_hook = pre_hook
        self._post_hook = post_hook

    def register(
        self, message_type: type[M], handler: Handler[M, Any] | Handlers[M, Any]
    ):
        self._handler_map[message_type] = handler

    @overload
    async def dispatch(self, message: Message) -> Any:
        ...

    @overload
    async def dispatch(
        self, message: Message, return_hooked_task: Literal[True] = True
    ) -> tuple[Any, asyncio.Future[list[Any]]]:
        ...

    async def dispatch(self, message: Message, return_hooked_task: bool = False):
        result, hooked = await asyncio.create_task(self._dispatch(message))
        coros = (self._dispatch(msg) for msg in hooked)
        hooked_task = asyncio.gather(*coros, return_exceptions=True)
        if return_hooked_task:
            return result, hooked_task
        await hooked_task
        return result

    async def _dispatch(self, message: Message):
        with MessageCatchContext():
            handler = self._handler_map[type(message)]
            if isinstance(handler, tuple):
                result = await handle_parallel(
                    message, handler, self._deps, self._pre_hook, self._post_hook
                )
            else:
                result = await handle(
                    message, handler, self._deps, self._pre_hook, self._post_hook
                )
            messages = get_issued_messages()
        return result, messages



from typing import Optional, Protocol, TypeVar


AggregateType = TypeVar("AggregateType")
IdentityType = TypeVar("IdentityType")


class CollectionOrientedRepositoryProtocol(Protocol[AggregateType, IdentityType]):
    async def add(self, _aggregate: AggregateType) -> None:
        ...

    async def get(self, _id: IdentityType) -> Optional[AggregateType]:
        ...

    async def delete(self, _aggregate: AggregateType) -> None:
        ...

    def next_identity(self) -> IdentityType:
        ...


class PersistenceOrientedRepositoryProtocol(Protocol[AggregateType, IdentityType]):
    async def save(self, _aggregate: AggregateType) -> None:
        ...

    async def get(self, _id: IdentityType) -> Optional[AggregateType]:
        ...

    async def delete(self, _aggregate: AggregateType) -> None:
        ...

    def next_identity(self) -> IdentityType:
        ...



from dataclasses import dataclass, field
from typing import Any, Callable, Optional
from sqlalchemy.ext.asyncio import AsyncSession

from ....port.out_.repository import (
    CollectionOrientedRepositoryProtocol,
    AggregateType,
    IdentityType,
)


@dataclass
class BaseSQLAlchemyRepository(
    CollectionOrientedRepositoryProtocol[AggregateType, IdentityType]
):

    AGGREGATE_TYPE: type[AggregateType] = field(init=False)
    IDENTITY_FACTORY: Callable[[], IdentityType] = field(init=False)

    _session: AsyncSession

    async def get(self, id: Any) -> Optional[AggregateType]:
        return await self._session.get(self.AGGREGATE_TYPE, id)  # type: ignore

    async def add(self, aggregate: AggregateType) -> None:
        self._session.add(aggregate)  # type: ignore

    async def delete(self, aggregate: AggregateType) -> None:
        await self._session.delete(aggregate)  # type: ignore

    def next_identity(self) -> IdentityType:
        return self.IDENTITY_FACTORY()



from typing import Any, Protocol, TypeVar


T_contra = TypeVar("T_contra", contravariant=True)
R_co = TypeVar("R_co", covariant=True)


class _Handler(Protocol[T_contra, R_co]):

    __name__: str

    async def __call__(self, _msg: T_contra, **_: Any) -> R_co:
        ...


Handler = _Handler


from types import TracebackType
from typing import Optional, Protocol
from typing_extensions import Self


class UnitOfWorkProtocol(Protocol):
    async def __aenter__(self) -> Self:
        ...

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        ...

    async def commit(self) -> None:
        ...

    async def rollback(self) -> None:
        ...
